 # !/usr/bin/env python
# coding: utf-8

##
# @file
# @brief Postprocessor class, implemented ZIDS_Postprocessor base.

from gaiaframework.base.pipeline.predictables.predictable import ZIDS_Predictable
from typing import List, Union

from gaiaframework.base.pipeline.postprocessor import ZIDS_Postprocessor
from ..artifacts.shared_artifacts import generatedProjectNameSharedArtifacts
from ..schema.outputs import generatedProjectNameOutputs


class generatedClass(ZIDS_Postprocessor):
    """generatedClass class (Postprocessor) implements ZIDS_Postprocessor base class.
    It is the last stage of the pipeline, its main focus is to return the results in the required format.
    """

    def __init__(self, artifacts: generatedProjectNameSharedArtifacts=None) -> None:
        """! generatedClass class (Postprocessor) initializer

        Args:
            artifacts(generatedProjectNameSharedArtifacts): Shared artifacts instance.
        """

        ##
        # @hidecallgraph @hidecallergraph
        super().__init__(artifacts)

    def config(self):
        """Implement here configurations required on Preprocess stage. Overrides ZIDS_Component.config()"""
        pass

    def normalize_output(self, predictables: Union[ZIDS_Predictable, List[ZIDS_Predictable]]) -> Union[generatedProjectNameOutputs, List[generatedProjectNameOutputs]]:
        """! Converts received predictable objects to generatedProjectNameOutputs datatype.

        Args:
            predictables: List[ZIDS_Predictable] - Predictable objects, the results from the model.

        Returns:
            generatedProjectNameOutputs: List[generatedProjectNameOutputs] - Results converted to Outputs format.
        """

        output: generatedProjectNameOutputs = ''
        isList = isinstance(predictables, list)
        if isList:
            output: List[generatedProjectNameOutputs] = []
        if isList:
            if predictables and len(predictables):
                for item in predictables:
                    output.append(self.get_output_object(item))
        else:
            output = self.get_output_object(predictables)
        return output

    def get_output_object(self, predictable):
        """! Parse a single predictable item, needs to be implemented.

        Args:
            predictable: ZIDS_Predictable - Single predictable object.

        Returns:
            generatedProjectNameOutputs: generatedProjectNameOutputs - Parsed results

        Raises:
            NotImplementedError

        """

        ##
        # Implementation example:
        # @code{.py}
        # prob = predictable[-1]
        # pred = False
        #
        # if prob > self.artifacts.threshold:
        #     pred = True
        # return generatedProjectNameOutputs(pred=pred, prob=prob)
        # @endcode

        raise NotImplementedError




from gaiaframework.base.cloud_eval.cloud_eval_client_base import CloudEvalClientBase
from tester.evaluator import generatedProjectNameEvaluator

UNIMPLEMENTED_ERROR_MESSAGE = "Please override the parent base class methods for using cloud eval"


class CloudEvalClient(CloudEvalClientBase):
    def __init__(self):
        self.model_evaluator = generatedProjectNameEvaluator()
    """
    Override the parent base class methods for using cloud eval
    """
    def get_request_headers(self, row: dict):
        raise NotImplementedError(UNIMPLEMENTED_ERROR_MESSAGE)

    def get_request_payload(self, row: dict):
        raise NotImplementedError(UNIMPLEMENTED_ERROR_MESSAGE)

    def extract_model_predictions_from_response(self, predictions):
        raise NotImplementedError(UNIMPLEMENTED_ERROR_MESSAGE)

    def get_endpoint_name(self):
        raise NotImplementedError(UNIMPLEMENTED_ERROR_MESSAGE)

    def evaluate_model_results(self, row: dict):
        raise NotImplementedError(UNIMPLEMENTED_ERROR_MESSAGE)

    def set_format_output(self, row: dict):
        raise NotImplementedError(UNIMPLEMENTED_ERROR_MESSAGE)



import requests
import json
import sys
from server.token_generator import tokenGenerator

env = sys.argv[1]

t = tokenGenerator()
jwtToken = t.generateToken(env.upper() + '_Conf')

main_model = 'original_project_name'
repo = 'gaia/' + main_model
data = ""
url = "/api/v1/models/getModelOptions"
if env == 'stg':
	url = "/api/v1/models/getModelOptions"
headers = {
    'x-token': 'fake-super-secret-token',
    'Authorization': 'Bearer ' + jwtToken
}

response = requests.post(url, json=data, headers=headers)
try:
	result = json.loads(response.content)
	models = result['data']['models']
	model_options = [{'name': model['name'], 'id': model['id'], 'repo': model['repo']} for model in models if model['repo'] == repo]
	if len(model_options) == 0:
		print(f'no model named {main_model} in dsp')
	else:
		if not len(model_options):
			print(f'cant find model name {main_model} in dsp')
		else:
			print(model_options)
except Exception as e:
	print('error', e)



import os
import asyncio
import concurrent.futures
import time
import dotenv
from .co_here import CoHereService
from .chat_gpt import ChatGptService

class VendorsManager:
    debug = False
    timer_start = 0
    timer_end = 0
    total_took = 0
    available_vendors = [ChatGptService.name, CoHereService.name]
    compared_vendors = []
    vendors_results = []
    chatGptService=None
    coHereService=None
    def __init__(self, vendors=None, env_file_path=None, api_keys: dict={}, debug=False):
        self.debug = debug
        self.load_env_file(env_file_path=env_file_path)
        chatgpt_key = (os.environ['OPEN_AI_KEY'] if 'OPEN_AI_KEY' in os.environ else (
            api_keys['OPEN_AI_KEY'] if 'OPEN_AI_KEY' in api_keys else ''))
        cohere_key = (os.environ['COHERE_AI_KEY'] if 'COHERE_AI_KEY' in os.environ else (
            api_keys['COHERE_AI_KEY'] if 'COHERE_AI_KEY' in api_keys else ''))
        self.chatGptService = ChatGptService(api_key=chatgpt_key)
        self.coHereService = CoHereService(api_key=cohere_key)
        if vendors is None:
            vendors = self.available_vendors
        self.compared_vendors = vendors

    def load_env_file(self, env_file_path):
        path = '.vendors_keys'
        if env_file_path:
            path = env_file_path
        dotenv_file = os.path.join(path)
        if os.path.isfile(dotenv_file):
            dotenv.load_dotenv(dotenv_file)
    def reset(self):
        self.vendors_results = []
        self.timer_start = 0
        self.timer_end = 0
        self.total_took = 0

    def run_cohere(self, prompt) -> dict:
        prompt_length = len(prompt)
        words_length = len(prompt.split(' '))
        max_tokens = 300 + (words_length * 2) * 3
        text = ''
        total_tokens = 0
        cost = 0
        try:
            prediction = self.coHereService.run_completion(prompt, max_tokens, False)
            text = prediction[0].text
            prompt_tokens = self.coHereService.get_tokens(prompt)
            prediction_tokens = self.coHereService.get_tokens(text)
            total_tokens = prompt_tokens + prediction_tokens
            cost = self.coHereService.get_price(total_tokens)
        except Exception as e:
            text = str(e)
        return {'text': text, 'total_tokens': total_tokens, 'cost': cost}

    def run_chat_gpt(self, prompt) -> dict:
        prompt_length = len(prompt)
        words_length = len(prompt.split(' '))
        max_tokens = 300 + (words_length * 2) * 3
        # prediction = chatGptService.run_completion(prompt, max_tokens, False)
        text = ''
        total_tokens = 0
        cost = 0
        try:
            # text = prediction.choices[0].text
            prediction = self.chatGptService.run_chat(prompt, [], False)
            text = prediction.choices[0].message.content
            prompt_tokens = total_tokens = self.chatGptService.get_tokens(prompt)
            prediction_tokens = total_tokens = self.chatGptService.get_tokens(text)
            total_tokens = prompt_tokens + prediction_tokens
            total_tokens = prediction.usage.total_tokens
            cost = self.chatGptService.get_price(total_tokens)
        except Exception as e:
            text = str(e)
        return {'text': text, 'total_tokens': total_tokens, 'cost': cost}

    async def upper_cased_task(self, vendor: str, prompt: str) -> dict:
        if vendor == ChatGptService.name:
            result = self.run_chat_gpt(prompt)
        if vendor == CoHereService.name:
            result = self.run_cohere(prompt)
        await asyncio.sleep(1)
        return result

    async def chat_gpt_task(self, prompt: str):
        if self.debug:
            print(f"Starting task1 for {prompt}")
        await asyncio.sleep(1)
        result = self.run_chat_gpt(prompt)
        if self.debug:
            print(f"Ending task1 for {prompt}")
        return result

    async def co_here_task(self,prompt: str):
        if self.debug:
            print(f"Starting task2 for {prompt}")
        await asyncio.sleep(1)
        result = self.run_cohere(prompt)
        if self.debug:
            print(f"Ending task2 for {prompt}")
        return result

    async def generic_task(self, vendor: str, prompt: str, index: int) -> dict:
        if self.debug:
            print(f"Starting generic_task in {vendor} for {prompt}")
        result = 'no vendor'
        with concurrent.futures.ThreadPoolExecutor() as executor:
            timer_start = time.perf_counter()
            if vendor == ChatGptService.name:
                result = await asyncio.get_running_loop().run_in_executor(executor, self.run_chat_gpt, prompt)
            if vendor == CoHereService.name:
                result = await asyncio.get_running_loop().run_in_executor(executor, self.run_cohere, prompt)
            timer_stop = time.perf_counter()
            # print('self.vendors_results', self.vendors_results)
            if len(self.vendors_results) and self.vendors_results[index]:
                self.vendors_results[index][f'{vendor}_took'] = f'{timer_stop - timer_start:0.4f}'
        if self.debug:
            print(f"Ending generic_task in {vendor} for {prompt}")
        return result

    async def process_item(self, item: str):
        tasks = []
        obj = {'input': item, 'took': ''}
        for vendor in self.compared_vendors:
            obj[vendor] = ''
            tasks.append(self.generic_task(vendor, item, -1))
        # tasks.append(self.chat_gpt_task(prompt))
        # tasks.append(self.co_here_task(prompt))
        timer_start = time.perf_counter()

        self.vendors_results.append(
            obj
        )
        results = await asyncio.gather(*tasks)
        timer_stop = time.perf_counter()
        if self.debug:
            print(f"Performed {item} on {self.compared_vendors}... {timer_stop - timer_start:0.4f} seconds")
        if self.vendors_results[-1]:
            for idx, vendor in enumerate(self.compared_vendors):
                self.vendors_results[-1][vendor] = results[idx]
            self.vendors_results[-1]['took'] = f'{timer_stop - timer_start:0.4f}'
        # print(f"Results for {item}: {result1}, {result2}")

    async def run_all_parallel(self, prompts: list):
        self.reset()
        self.timer_start = time.perf_counter()
        await self.process_items(prompts)
        self.timer_end = time.perf_counter()
        self.total_took = self.timer_end - self.timer_start

    async def process_items(self, prompts: list):
        tasks = []
        count = 0
        for prompt in prompts:
            obj = {'input': prompt}
            for vendor in self.compared_vendors:
                obj[vendor] = ''
                tasks.append(self.generic_task(vendor, prompt, count))
            self.vendors_results.append(obj)
            count += 1
        # tasks.append(self.chat_gpt_task(prompt))
        # tasks.append(self.co_here_task(prompt))
        timer_start = time.perf_counter()

        # self.vendors_results.append(
        #     obj
        # )
        results = await asyncio.gather(*tasks)
        timer_stop = time.perf_counter()
        if self.debug:
            print(f"Performed all prompts on {self.compared_vendors}... {timer_stop - timer_start:0.4f} seconds")
        # print('results', results)
        # print('self.vendors_results', self.vendors_results)
        count = 0
        if len(self.compared_vendors):
            for idx in range(0, len(results), len(self.compared_vendors)):
                if self.vendors_results[count]:
                    for i, vendor in enumerate(self.compared_vendors):
                        self.vendors_results[count][vendor] = results[idx + i]
                    # self.vendors_results[count]['took'] = f'{timer_stop - timer_start:0.4f}'
                    count += 1

    async def run(self, prompts: list):
        self.reset()
        self.timer_start = time.perf_counter()
        for item in prompts:
            await self.process_item(item)
        self.timer_end = time.perf_counter()
        self.total_took = self.timer_end - self.timer_start



import json
import ast

import click
import os, shutil
import re
import sys
import platform
import subprocess

isWindows = platform.system() == 'Windows'

__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(__file__)))

directories = {
    'main': '',
    'pipeline': '',
    'artifacts': '',
    'models': '',
    'vocabs': '',
    'other': '',
    'preprocessor': '',
    'predictables': '',
    'predictors': '',
    'forcers': '',
    'postprocessor': '',
    'schema': '',
    'tester': '',
    'test_schema': '',
    'trainer': '',
    'server': '',
    'dsp': '',
}


class AliasedGroup(click.Group):
    def get_command(self, ctx, cmd_name):
        try:
            cmd_name = ALIASES[cmd_name].name
        except KeyError:
            pass
        return super().get_command(ctx, cmd_name)

@click.group(cls=AliasedGroup)
@click.version_option(package_name='gaiaFramework')
def cli():
    """
    DS framework cli

    ** how to use **

    g = generate

    p = project

    create project

    gaia-cli g project my-new-project

    gaia-cli g p my-new-project

    gaia-cli generate project my-new-project

    gaia-cli generate p my-new-project

    cd my-new-project

    create forcer

    gaia-cli g forcer my-new-forcer

    create predictable

    gaia-cli g predictable my-new-predictable

    create tester files

    gaia-cli create-tester-files

    run server

    gaia-cli run-server

    create deploy files

    gaia-cli create-deploy-files

    create cloud eval files

    gaia-cli create-cloud-eval-files

    run evaluation on a csv input

    gaia-cli evaluate input-csv-file batch-size

    """
# @click.option("--type", prompt="type", help="type of component")
# @click.option("--project_name", prompt="project_name", help="project_name")
# def apis(type, project_name):


@cli.command()
@click.argument('gen_type')
@click.argument('project_name')
def generate(gen_type, project_name):
    """List all cataloged APIs."""
    try:
        if gen_type in FUNC_ALIASES:
            gen_type = FUNC_ALIASES[gen_type]
        f = globals()["generate_%s" % gen_type]
    except Exception as e:
        click.echo('type ' + gen_type + ' not found')
        return
    f(project_name)


@cli.command()
def run_server():
    currentPipelineFolder = os.path.abspath(os.getcwd())
    currentParentFolder = os.path.join(currentPipelineFolder,"..")
    os.environ["PYTHONPATH"] = currentPipelineFolder
    folder = currentPipelineFolder + '/server/main.py'
    subprocess.call('python ' + folder, shell=True)

@cli.command()
@click.argument('tag')
@click.argument('env', default="prd")
def cloud_eval(tag, env):
    currentPipelineFolder = os.path.abspath(os.getcwd())
    currentParentFolder = os.path.join(currentPipelineFolder,"..")
    os.environ["PYTHONPATH"] = currentPipelineFolder
    folder = currentPipelineFolder + '/tester/dsp/get_model_options.py'
    output = subprocess.check_output('python ' + folder + ' ' + env, shell=True, encoding='utf-8').strip()
    if output:
        # print('output', output)
        l = output.split('\n')
        m = l[-1]
        try:
            l = ast.literal_eval(m)
            o = l[0]
            model = {
                'name': o['name'],
                'id': o['id'],
                'repo': o['repo'],
            }
            folder = currentPipelineFolder + '/tester/dsp/get_dataset_options.py'
            output = subprocess.check_output('python ' + folder + ' ' + env + ' ' + model['name'] + ' ' + str(model['id']),
                                             shell=True, encoding='utf-8').strip()
            l = output.split('\n')
            m = l[-1]
            l = ast.literal_eval(m)
            selected = let_user_pick(l)
            # print('selected', selected)
            if not selected == None:
                selected_dataset = l[selected]
                # print('model',model)
                # print('selected_dataset', selected_dataset)
                folder = currentPipelineFolder + '/tester/dsp/add_experiment.py'
                subprocess.call(('dvc_push.sh'), shell=True)
                output = subprocess.call(('python', folder, env, json.dumps(model), json.dumps(selected_dataset), tag), shell=True)
                print('you can see your experiment in que')
            else:
                print('selection does not exist')
        except Exception as e:
            print(m)
    else:
        print('could not get models')


@cli.command()
def create_deploy_files():
    copy_deploy_files('')


@cli.command()
def create_cloud_eval_files():
    currentPipelineFolder = os.path.abspath(os.getcwd())
    copy_cloud_eval_files(currentPipelineFolder)


@cli.command()
def create_tester_files():
    create_tester_env()


@cli.command()
@click.argument('csv_file_path')
@click.argument('batch_size')
def evaluate(csv_file_path, batch_size):
    if os.path.isfile(csv_file_path):
        current_pipeline_folder = os.path.abspath(os.getcwd())
        os.environ["PYTHONPATH"] = current_pipeline_folder
        folder = current_pipeline_folder + '/tester/general_tester.py'
        subprocess.call('python ' + folder + ' ' + csv_file_path + ' ' + batch_size, shell=True)
    else:
        click.echo('file: ' + csv_file_path + ' not found')


ALIASES = {
    "g": generate
}

FUNC_ALIASES = {
    "p": "project"
}


def generate_project(project_name):
    # project_name = clean_name(project_name)
    click.echo('Generating project: ' + project_name)
    create_project(project_name)


def generate_forcer(fileName):
    fileName = clean_name(fileName)
    create_exist_pipeline_file('forcer', fileName)


def generate_predictable(fileName):
    fileName = clean_name(fileName)
    create_exist_pipeline_file('predictable', fileName)


def clean_name(name):
    name = name.replace('-', '_')
    return name


def create_folders(project_name):
    global directories
    directories['main'] = project_name
    if not os.path.exists(directories['main']):
        os.mkdir(directories['main'])

    create_main_folders('config', 'main')
    create_main_folders('pipeline', 'main')
    create_main_folders('artifacts', 'pipeline')
    create_main_folders('models', 'artifacts')
    create_main_folders('vocabs', 'artifacts')
    # create_main_folders('other', 'artifacts')
    create_main_folders('preprocessor', 'pipeline')
    create_main_folders('predictables', 'pipeline')
    create_main_folders('predictors', 'pipeline')
    create_main_folders('forcers', 'pipeline')
    create_main_folders('postprocessor', 'pipeline')
    create_main_folders('schema', 'pipeline')
    create_main_folders('tester', 'main')
    create_main_folders('test_schema', 'tester')
    create_main_folders('dsp', 'tester')
    create_main_folders('trainer', 'main')
    create_main_folders('server', 'main')


def create_project(project_name):
    create_folders(project_name)
    original_project_name = project_name
    project_name = clean_name(project_name)

    # Create pipeline files
    create_pipeline_file(project_name, directories['artifacts'], 'shared_artifacts')
    create_pipeline_file(project_name, directories['preprocessor'], 'preprocess')
    create_pipeline_file(project_name, directories['predictors'], 'predictor')
    create_pipeline_file(project_name, directories['forcers'], 'forcer')
    create_pipeline_file(project_name, directories['postprocessor'], 'postprocess')
    create_pipeline_file(project_name, directories['predictables'], 'predictable')
    create_pipeline_file(project_name, directories['pipeline'], 'pipeline')
    create_pipeline_file(project_name, directories['main'], 'pipeline_test', False)

    # Create schema files
    create_schema_file(project_name, directories['schema'], 'inputs')
    create_schema_file(project_name, directories['schema'], 'outputs')
    create_schema_file(project_name, directories['schema'], '__init__', False)

    # Create tester files
    create_tester_file(project_name, directories['tester'], 'general_tester', False)
    create_tester_file(project_name, directories['tester'], 'evaluator', False)
    create_schema_file(project_name, directories['test_schema'], 'test_input', False)
    create_schema_file(project_name, directories['test_schema'], 'test_output', False)

    create_dsp_file(original_project_name, directories['dsp'], 'get_model_options', False)
    create_dsp_file(original_project_name, directories['dsp'], 'get_dataset_options', False)
    create_dsp_file(original_project_name, directories['dsp'], 'add_experiment', False)

    # Create server files
    create_server_file(project_name, directories['server'], 'main', False)
    create_server_file(project_name, directories['server'], 'test_server_post', False)
    create_server_file(project_name, directories['server'], 'test_server_post_stream', False)
    create_server_file(project_name, directories['server'], 'pool', False)
    create_server_file(project_name, directories['server'], 'token_generator', False)
    create_server_file(project_name, directories['server'], '__init__', False)

    create_project_config_json()
    create_project_gitignore()
    create_server_config_json()
    copy_deploy_files(directories['main'])
    copy_cloud_eval_files(directories['main'])

    change_to_project_dir()
    run_dvc_init()
    run_git_init()


def create_main_folders(targetDir, baseDir):
    global directories
    directories[targetDir] = directories[baseDir] + '/' + targetDir
    if not os.path.exists(directories[targetDir]):
        os.mkdir(directories[targetDir])


def create_pipeline_file(project_name, folder, pipelineType, createInitFile=True):
    data = read_template_file(pipelineType)
    replace_in_template_and_create_file(project_name, folder, pipelineType, data, createInitFile)


def create_schema_file(project_name, folder, pipelineType, createInitFile=True):
    data = read_template_file('schema/' + pipelineType)
    replace_in_template_and_create_file(project_name, folder, pipelineType, data, createInitFile)


def create_server_file(project_name, folder, pipelineType, createInitFile=True):
    data = read_template_file('tester/server/' + pipelineType)
    replace_in_template_and_create_file(project_name, folder, pipelineType, data, createInitFile)


def create_tester_file(project_name, folder, test_file_type, createInitFile=True):
    data = read_template_file('tester/' + test_file_type)
    replace_in_template_and_create_file(project_name, folder, test_file_type, data, createInitFile)

def create_dsp_file(project_name, folder, test_file_type, createInitFile=True):
    data = read_template_file('tester/dsp/' + test_file_type)
    replace_in_template_and_create_file(project_name, folder, test_file_type, data, createInitFile)


def create_exist_pipeline_file(type, fileName):
    pipelineType = type + 's'
    folder = 'pipeline/' + pipelineType
    if os.path.exists(folder):
        data = read_template_file(type)
        fileNameNoUnderscore = to_capitalize_no_underscore(fileName)
        className = fileNameNoUnderscore + type.capitalize()
        currentPipelineFolder = os.path.basename(os.getcwd())
        currentDir = folder.replace('/', '.')

        data = data.replace('generatedClass', className)

        new_file = folder + "/" + fileName + ".py"
        current_init_file = folder + "/__init__.py"
        new_init_export = "from " + '.' + fileName + " import " + className

        create_file(new_file, data)
        create_init_file(current_init_file, new_init_export)
        inject_to_pipeline(fileName, type, className, new_init_export)
    else:
        print('please create a project and go to project location first')
    pass


def create_tester_env():
    """
    This function will create a testing environment, if one does not already exist in the project.
    It will create the testing, evaluation, and relevant schema files.
    """
    project_name = os.getcwd().split('/')[-1]
    dst_folder = 'tester/'
    schema_dst_folder = 'tester/test_schema/'
    tester_file_list = ["tester/general_tester",
                        "tester/evaluator",
                        "schema/test_input",
                        "schema/test_output"]

    if not os.path.exists(dst_folder):
        os.mkdir(dst_folder)
        os.mkdir(schema_dst_folder)

        for filename_with_path in tester_file_list:
            file_only = filename_with_path.split('/')[-1]
            if file_only.startswith('test'):
                dst_folder = schema_dst_folder

            data = read_template_file(filename_with_path)

            clean_project_name = clean_existing_project_name(project_name)
            filename_no_underscore = to_capitalize_no_underscore(file_only)

            class_name = clean_project_name + filename_no_underscore

            data = data.replace('generatedClass', class_name)
            data = data.replace('generatedProjectName', clean_project_name)

            new_file = dst_folder + "/" + file_only + ".py"
            create_file(new_file, data)

            # Test schema does not use generated class name for now, so this code should be modified before uncommenting
            # current_init_file = dst_folder + "/" + "__init__.py"
            # new_init_export = "from " + dst_folder.replace('/', '.') + "." + file_only + " import " + class_name
            # create_init_file(current_init_file, new_init_export)
    else:
        print('Tester folder Already exists. If you wish to recreate, please erase folder and retry')
    pass


def read_template_file(filename):
    with open(os.path.join(__location__, 'gaiaframework/cli/' + filename + '_template.py'), 'r') as file:
        data = file.read()
        return data


def read_file(file_path):
    with open(file_path, 'r') as file:
        data = file.read()
        return data


def write_to_file(file_path, data):
    with open(file_path, 'w') as file:
        file.write(data)
        file.close()


def replace_in_template_and_create_file(project_name, folder, pipelineType, data, createInitFile):
    pipeline_type_no_underscore = to_capitalize_no_underscore(pipelineType)
    project_name_no_underscore = to_capitalize_no_underscore(project_name)
    className = project_name_no_underscore + pipeline_type_no_underscore
    classNameForBaseObject = project_name_no_underscore + pipeline_type_no_underscore
    currentDir = folder.replace('/', '.')
    currentDirNoMainDir = folder.replace(directories['main'] + '/', '').replace('/', '.')
    currentBaseDir = directories['main'].replace('/', '.')
    currentPipelineDir = directories['pipeline'].replace('/', '.')

    data = data.replace('generatedClassName', classNameForBaseObject)
    data = data.replace('generatedClass', className)
    data = data.replace('original_project_name', project_name)
    data = data.replace('generatedProjectName', project_name_no_underscore)
    data = data.replace('generatedDirectory', currentDir)
    data = data.replace('generatedBaseDir', currentBaseDir)
    data = data.replace('generatedPipelineDir', currentPipelineDir)

    new_file = folder + "/" + pipelineType + ".py"
    create_file(new_file, data)

    if createInitFile:
        new_init_file = folder + "/__init__.py"
        new_init_export = "from " + '.' + pipelineType + " import " + className
        create_init_file(new_init_file, new_init_export)


def create_predictable_file(fileName):
    pass


def create_file(new_file_path, data):
    if not os.path.exists(new_file_path):
        f = open(new_file_path, "w")
        f.write(data)
        f.close()


def create_init_file(init_path, init_export):
    if not os.path.exists(init_path):
        f = open(init_path, "w")
        f.write(init_export)
        f.close()
    else:
        f = open(init_path, 'r+')
        data = f.read()
        if init_export not in data:
            if len(data) and not data.endswith('\n'):
                f.write('\n')
            f.write(init_export)
            f.close()


def inject_to_pipeline(fileName, type, className, new_init_export):
    file_path = 'pipeline/pipeline.py'
    if os.path.exists(file_path):
        data = read_file(file_path)
        new_tab_line = '\n'
        data_changed = False
        last_index_of_import = -1
        first_index_of_add_component = -1
        index_of_class = re.search(r'class[^\n]*', data)

        # finding current indent config
        index_of_build_pipeline = re.search(r'def build_pipeline[^\n]*', data)
        index_of_build_preprocessor = re.search(r'self.preprocessor =[^\n]*', data)
        index_of_build_postprocessor = re.search(r'self.postprocessor =[^\n]*', data)
        index_of_build_pipeline_row = re.search(r'.*def build_pipeline[^\n]*', data)
        if index_of_build_pipeline:
            build_pipeline_indent = (index_of_build_pipeline.start() - index_of_build_pipeline_row.start()) * 2
            new_tab_line = new_tab_line.ljust(build_pipeline_indent + 1)

        attr = fileName + type.title()
        new_component_line = 'self.' + attr + ' = ' + className + '()'
        add_component_line = 'self.add_component(self.' + attr + ')'

        # finding imports and add components indexes
        last_index_of_add_component = -1
        all_from_import = [i.end() for i in re.finditer(r'from[^\n]*', data)]
        all_add_components = [[i.start(), i.end()] for i in re.finditer(r'self.add_component[^\n]*', data)]
        if len(all_from_import):
            last_index_of_import = all_from_import[-1]
        if len(all_add_components):
            first_index_of_add_component = all_add_components[0][0]
            last_index_of_add_component = all_add_components[-1][-1]
        # finding imports and add components indexes


        index_to_add = 0

        # add import to end of imports or to top of file
        if last_index_of_import > -1 and new_init_export not in data:
            s = '\n' + new_init_export
            index_to_add += len(s)
            data = data[:last_index_of_import] + s + data[last_index_of_import:]
            data_changed = True
        elif index_of_class and new_init_export not in data:
            s = new_init_export + '\n\n'
            index_to_add += len(s)
            data = data[:index_of_class.start()] + s + data[index_of_class.start():]
            data_changed = True

        # check if build_pipeline exist but with no components yet
        if first_index_of_add_component == -1 and last_index_of_add_component == -1 and index_of_build_pipeline:
            s = new_tab_line
            current_end = index_of_build_pipeline.end()
            if index_of_build_preprocessor:
                current_end = index_of_build_preprocessor.end()
            if index_of_build_postprocessor:
                current_end = index_of_build_postprocessor.end()
            index = current_end + index_to_add
            index_to_add += len(s)
            data = data[:index] + s + data[index:]
            data_changed = True
            first_index_of_add_component = current_end
            last_index_of_add_component = current_end

        # adding new component line
        if first_index_of_add_component > -1 and new_component_line not in data:
            first_index_of_add_component += index_to_add
            if len(all_add_components):
                new_component_line = new_component_line + new_tab_line
            index_to_add += len(new_component_line)
            data = data[:first_index_of_add_component] + new_component_line + data[first_index_of_add_component:]
            data_changed = True

        #adding add_component line
        if last_index_of_add_component > -1 and add_component_line not in data:
            last_index_of_add_component += index_to_add
            add_component_line = new_tab_line + add_component_line
            index_to_add += len(add_component_line)
            data = data[:last_index_of_add_component] + add_component_line + data[last_index_of_add_component:]
            data_changed = True

        if data_changed:
            write_to_file(file_path, data)


def create_project_config_yaml():
    with open(os.path.join(__location__, 'gaiaframework/cli/config.yaml'), 'r') as file:
        data = file.read()
        data = data.replace('generatedDirectory', directories['main'])
        new_file = directories['main'] + '/pipeline/config.yaml'
        if not os.path.exists(new_file):
            f = open(new_file, "w")
            f.write(data)
            f.close()


def create_project_config_json():
    with open(os.path.join(__location__, 'gaiaframework/cli/config.json'), 'r') as file:
        data = file.read()
        data = data.replace('generatedDirectory', directories['main'])
        new_file = directories['main'] + '/config/config.json'
        if not os.path.exists(new_file):
            f = open(new_file, "w")
            f.write(data)
            f.close()


def create_project_gitignore():
    with open(os.path.join(__location__, 'gaiaframework/cli/.gitignore'), 'r') as file:
        data = file.read()
        new_file = directories['main'] + '/.gitignore'
        if not os.path.exists(new_file):
            f = open(new_file, "w")
            f.write(data)
            f.close()


def create_server_config_json():
    with open(os.path.join(__location__, 'gaiaframework/cli/cors_allowed_origins.json'), 'r') as file:
        data = file.read()
        data = data.replace('generatedDirectory', directories['main'])
        new_file = directories['server'] + '/cors_allowed_origins.json'
        if not os.path.exists(new_file):
            f = open(new_file, "w")
            f.write(data)
            f.close()


def change_to_project_dir():
    os.chdir(directories['main'])


def run_dvc_init():
    dir_path = os.getcwd()
    if not os.path.isdir(dir_path + '/.dvc'):
        command = 'dvc_init.sh'
        if not isWindows:
            command = './' + command
        subprocess.call(command, shell=True)


def run_git_init():
    dir_path = os.getcwd()
    if not os.path.isdir(dir_path + '/.git'):
        command = 'git_init.sh'
        if not isWindows:
            command = './' + command
        subprocess.call(command, shell=True)


def copy_deploy_files(main_dir):
    currentPipelineFolder = os.path.basename(os.getcwd())
    if main_dir:
        currentPipelineFolder = main_dir
    dir = os.path.join(__location__, 'gaiaframework/cli/tester/deploy_files/')
    listOfFiles = list()
    for (dirpath, dirname, filenames) in os.walk(dir):
        dirpath = dirpath.replace(dir, '')
        listOfFiles += [os.path.join(dirpath, file) for file in filenames]
        for file_path in listOfFiles:
            with open(os.path.join(dir, file_path), 'r', encoding="utf8") as file:
                try:
                    data = file.read()
                    data = data.replace('{name-your-service}', currentPipelineFolder)
                    data = data.replace('{name-your-artifacts}', currentPipelineFolder)
                    dirToCreate = ''
                    if main_dir:
                        dirToCreate = main_dir + '/'
                    if not os.path.exists(dirToCreate + 'deploy'):
                        os.makedirs(dirToCreate + 'deploy')
                    if not os.path.exists(dirToCreate + file_path):
                        f = open(dirToCreate + file_path, "w")
                        f.write(data)
                        f.close()
                        # Make sure that .sh files can be executed
                        if (dirToCreate + file_path).endswith('.sh'):
                            os.chmod(dirToCreate + file_path, 0o744)
                except Exception as ex:
                    pass
    # for item in os.listdir(dir):
    #     s = os.path.join(dir, item)
    #     d = os.path.join(main_dir, item)
    #     if os.path.isdir(s):
    #         shutil.copytree(s, d, False, None)
    #     else:
    #         shutil.copy2(s, d)


def copy_cloud_eval_files(main_dir):
    currentPipelineFolder = os.path.basename(os.getcwd())
    if main_dir:
        currentPipelineFolder = main_dir
    service_folder = str(currentPipelineFolder)
    service_name = service_folder.split('/')[-1]
    project_name_no_underscore = clean_existing_project_name(service_name)
    dir = os.path.join(__location__, 'gaiaframework/cli/tester/cloud_eval/')
    listOfFiles = list()
    conflicted_files = []
    for (dirpath, dirname, filenames) in os.walk(dir):
        if "__pycache__" in dirpath:
            continue
        dirpath = dirpath.replace(dir, '')
        listOfFiles += [os.path.join(dirpath, file) for file in filenames]
        for file_path in listOfFiles:
            framework_source_file = os.path.join(dir, file_path)
            with open(framework_source_file, 'r', encoding="utf8") as file:
                try:
                    data = file.read()
                    data = data.replace('{name-your-service}', service_name)
                    data = data.replace('{name-your-artifacts}', service_name)
                    data = data.replace('generatedProjectName', project_name_no_underscore)
                    dirToCreate = ''
                    if main_dir:
                        dirToCreate = main_dir + '/cloud_eval/'
                    if not os.path.exists(dirToCreate):
                        os.makedirs(dirToCreate)
                    if "#cli_file_dest=[" in data:
                        start_ix = data.index("#cli_file_dest=[") + len("#cli_file_dest=[")
                        end_ix = data.index("]", start_ix)
                        destination_dir = data[start_ix:end_ix]
                        data = data[:data.index("#cli_file_dest=[")] + data[end_ix+1:]
                        dirToCreate = main_dir + '/' + destination_dir

                    if not os.path.exists(dirToCreate + file_path):
                        f = open(dirToCreate + file_path, "w")
                        f.write(data)
                        f.close()
                        # Make sure that .sh files can be executed
                        if (dirToCreate + file_path).endswith('.sh'):
                            os.chmod(dirToCreate + file_path, 0o744)
                    else:
                        with open(dirToCreate + file_path, "r") as f:
                            existing_data = f.read()
                        if existing_data != data:
                            conflicted_file = f"{dirToCreate + file_path}"
                            conflicted_files.append((conflicted_file, str(framework_source_file)))

                except Exception as ex:
                    pass

    if conflicted_files:
        print("The following files already exists, so not overwriting them. "
              "However, notice it is different from the framework version in")
        for conflicted_file in conflicted_files:
            print(f"Existing: [{conflicted_file[0]}] \t\t Framework source file [{conflicted_file[1]}]")


def to_capitalize_no_underscore(text):
    return ''.join(elem.capitalize() for elem in text.split('_'))


def clean_existing_project_name(text):
    no_underscore_text = to_capitalize_no_underscore(text)
    return ''.join(elem.capitalize() for elem in no_underscore_text.split('-'))

def let_user_pick(options):
    print("Please choose:")
    for idx, element in enumerate(options):
        print("{}) {}".format(idx+1,element))
    i = input("Enter number: ")
    try:
        if 0 < int(i) <= len(options):
            return int(i)-1
    except:
        pass
    return None


if __name__ == '__main__':
    cli(prog_name='cli')



"""Simple Pool object"""

from queue import Queue

"""
Adapted from http://pythonwise.blogspot.com/2016/09/simple-object-pools.html
"""
class InstancePool:
    """Pool of objects"""
    def __init__(self, objects):
        self._queue = Queue()
        for obj in objects:
            self._queue.put(obj)

    def lease(self):
        """Lease an object from the pool, should be used as contect manger. e.g.:

            with pool.lease() as conn:
                cur = conn.cursor()
                cur.execute('SELECT ...')
        """
        return self._queue.get()

    def _put(self, obj):
        self._queue.put(obj)





# from abc import ABCMeta, abstractmethod

class BaseVendor:
    # __metaclass__ = ABCMeta
    # @abstractmethod
    def __init__(self): raise NotImplementedError
    def run_chat(self): raise NotImplementedError
    def run_completion(self): raise NotImplementedError
    def get_tokens(self): raise NotImplementedError
    def get_price(self): raise NotImplementedError
    def calculate_price(self): raise NotImplementedError


from gaiaframework.utils.regex_handler import RegexHandler
from generatedMainDirectory.pipeline import generatedProjectNamePipeline
from gaiaframework.testable.tester_base import TesterBase
# from generatedMainDirectory.tester import generatedProjectNameDataset
from generatedMainDirectory.tester import generatedProjectNameReporter
import pandas as pd
import string
from tqdm import tqdm
import os
import json
import copy
import numpy as np


class generatedClass(TesterBase):
    @staticmethod
    def get_defaults():
        cfg = {}
        return cfg

    def __init__(self, name:str="generatedClassName",  **kwargs):
        TesterBase.__init__(self, name, generatedClass.get_defaults(),  **kwargs)
        self.ready = False

    def build(self, path:str='', load_samples=False):
        path, _ = self.convert_dataset_to_seg_format(path)
        self.init_pipeline()
        # self.init_dataset(path=path, load_samples=load_samples)
        self.init_reporter()
        self.ready = True

    def init_pipeline(self, *args, **kwargs):
        self.pipeline = generatedProjectNamePipeline.getInstance()

    # def init_dataset(self, path='',  do_convert=False, load_samples=False, *args, **kwargs):
    #     self.dataset = generatedProjectNameDataset(path=path, load_samples=load_samples)

    def init_reporter(self, path='', report_header='Signature Parser Analysis Report'):
        self.set_report_header(report_header)
        self.set_report_path(path)
        self.reporter = generatedProjectNameReporter()

    def parse(self, signature = ''):
        return self.pipeline(signature)

    def test(self, prediction_path = None, **kwargs):
        self.pre_test(**kwargs)
        if prediction_path:
            _, pred_df = self.convert_dataset_to_seg_format(prediction_path)
            self.dataset.set_df_prediction(pred_df)
        else:
            self.dataset.set_pipeline_prediction()
        report_col_names = self.get_report_df_cols()
        self.report_col_names = report_col_names
        self.report_df = pd.DataFrame(columns=report_col_names)
        for sample in tqdm(self.dataset.samples):
            sample_evaluations = self.evaluate(sample)
            if sample_evaluations:
                self.report_df.loc[len(self.report_df.index)] = sample_evaluations
        if self.reporter is not None:
            self.reporter(self.report_df, report_path=self.report_path, report_header=self.report_header)

    def pre_test(self):
        if not len(self.dataset.samples):
            self.dataset.populate()

    def get_report_df_cols(self):
        # def get_cols_per_attr(attr):
        #     return [f'{attr}_pred', f'{attr}_target', f'{attr}_prob', f'{attr}_tp',  f'{attr}_fp',  f'{attr}_tn', f'{attr}_fn']
        #
        # cols = []
        # cols+= ['name_prefix', 'name_first', 'name_middle', 'name_last', 'name_suffix', 'name_credentials']
        # cols+= ['tc_title', 'tc_company']
        # cols+= ['connections_phones', 'connections_mobiles', 'connections_faxes', 'connections_urls', 'connections_emails']
        # cols+= ['location_address', 'location_city', 'location_state', 'location_zip', 'location_country']
        #
        # self.report_base_cols = cols
        # cols_per_attr = [get_cols_per_attr(attr) for attr in cols]
        # flatten = lambda l: [item for sublist in l for item in sublist]
        # cols_as_list = flatten(cols_per_attr)
        # report_cols = ['text', 'pred', 'target'] + cols_as_list
        # return report_cols
        raise NotImplementedError

    def evaluate(self, sample):
        # def get_metrics(pred='', target=''):
        #     TP = 1*all([len(pred), len(target), pred==target])
        #     FP = 1*all([len(pred), pred!=target])
        #     TN = 1*all([not len(pred), not len(target), pred==target])
        #     FN = 1*all([not len(pred), len(target), pred!=target])
        #     return [TP, FP, TN, FN]
        #
        # text   = sample.x
        # pred   = sample.pred
        # target = sample.y
        #
        # pred_copy = copy.deepcopy(pred)
        # target_copy = copy.deepcopy(target)
        #
        # connections_pred   = self.convert_lists_to_str(pred_copy.get('connections', {}))
        # connections_target = self.convert_lists_to_str(target_copy.get('connections', {}))
        #
        # pred_copy['connections'] = connections_pred
        # target_copy['connections'] = connections_target
        #
        # # pred['connections'] = connections_pred
        # # target['connections'] = connections_target
        #
        # preds_dict  = self.F.flatten(pred)
        # target_dict = self.F.flatten(target)
        # pred_copy_dict  = self.F.flatten(pred_copy)
        # target_copy_dict = self.F.flatten(target_copy)
        #
        # # if not self.is_valid_signature(target_dict):
        # #     return None
        # row = [text, pred, target]
        #
        # if not self.report_base_cols:
        #     self.report_col_names = self.get_report_df_cols()
        #
        # for key in self.report_base_cols:
        #     prob       = -1
        #     pred_val   = preds_dict.get(key, '')
        #     target_val = target_dict.get(key, '')
        #     pred_val_c   = self.normalize_val_for_compare(pred_copy_dict.get(key, ''))
        #     target_val_c = self.normalize_val_for_compare(target_copy_dict.get(key, ''))
        #     metrices     = [pred_val, target_val, prob] + get_metrics(pred_val_c, target_val_c)
        #     row+=metrices
        #
        # return row
        raise NotImplementedError

    def convert_lists_to_str(self, d={}):
        # def list_to_str(k, v):
        #     val = [RegexHandler.remove_punct(str(t)).lower().replace(' ','') for t in v]
        #     val = list(set(val))
        #     val.sort()
        #     s = ''.join(val).strip()
        #     s = ''.join(list(set(s)))
        #     s = ''.join(sorted(s))
        #     if 'phone' in k :
        #         s = ''.join([c for c in s if c.isdigit()])
        #     else:
        #         s=s.replace('w','') # to ensure no www is considered + no order matters and no duplicates
        #     return s
        #
        # for k, v in d.items():
        #     if type(v) is list:
        #         d[k] = list_to_str(k, v)
        #
        # return d
        raise NotImplementedError

    def is_valid_signature(self, target_dict):
        company_exist    = len(target_dict.get('tc_company', ''))>0
        first_name_exist = len(target_dict.get('name_first', ''))>0
        last_name_exist  = len(target_dict.get('name_last', ''))>0
        return any([company_exist ,all([first_name_exist, last_name_exist])])

    def normalize_val_for_compare(self, val=''):
        if not val:
            return ''
        if not type(val) is str:
            val = str(val).lower()
        val = val.replace('\n', ' ')
        # s = RegexHandler.remove_punct(val, middle=False).lower()
        s = RegexHandler.remove_punct(val).lower()
        s = ''.join(filter(lambda x: x in string.printable, s)).strip()
        s = s.replace(' ','').replace('0', '')
        return s

    def convert_dataset_to_seg_format(self, path):
        """
        Convert dataset file from expanded format (each attribute as a column) to a segment format. Segment format has
        at least 2 columns: signature and segments. signature is the raw signature text column and segments is a string
        representation of the embedded attributes under the meta attributes: name, tc, connections and locations. In
        case the dataset is in expended format, the dataset is converted and saved locally and the new path is returned.
        In case the dataset file is already in segment format, its path is returned.
        @param path: path to dataset file
        @return: path to dataset file in segment format
        """
        df = pd.read_csv(path)
        if 'signature' in df.columns and 'segments' in df.columns:
            return path, df
        else:  # convert to segment format
            df_new = pd.DataFrame()
            df = df.replace({np.nan: ''})
            df_new['signature'] = df.signature
            df_new['segments'] = [json.dumps(self.convert_gt_to_seg(df.iloc[i])) for i in range(len(df))]
            path_new = os.path.join(os.path.dirname(path), os.path.basename(path).replace('.csv', '_seg_format.csv'))
            df_new.to_csv(path_new, index=False)
            return path_new, df_new

    @staticmethod
    def convert_gt_to_seg(entry):
        """
        Convert a dataframe entry into embedded segment attribute dictionary format.
        @param entry: dataframe entry
        @return: dictionary representation of embedded segment format.
        """
        d = {'name': {'first': entry['first'],
                      'last': entry['last'],
                      'middle': entry['middle'],
                      'prefix': entry['prefix'],
                      'suffix': entry['suffix'],
                      'credentials': entry['credentials']},
             'tc': {
                 'title': entry['title'],
                 'company': entry['company']
             },
             'connections': {
                 'emails': [entry['emails']],
                 'phones': [entry['phones']],
                 'faxes': [entry['faxes']],
                 'mobiles': [entry['mobiles']],
                 'urls': [entry['urls']],
             },
             'location': {
                 'address': entry['address'],
                 'city': entry['city'],
                 'state': entry['state'],
                 'zip': entry['zip'],
                 'country': entry['country'],
             }
             }
        return d

    @staticmethod
    def clear_df_before_saving(report_df):
        """
        Helper function clearing illegal characters in dataframe before saving as csv.
        @param report_df: a report dataframe to be stripped of illegal characters.
        @return: cleared dataframe (note some characters may be lost in the process.
        """
        for col in report_df.columns:
            if report_df[col].dtype == object and col.split('_')[-1] not in ['tp', 'tn', 'fp', 'fn']:
                report_df[col] = report_df[col].apply(
                    lambda x: np.nan if x == np.nan else str(x).encode('utf-8', 'replace').decode('utf-8'))
        return report_df

    @staticmethod
    def get_f1_precision_recall(var_name, df):
        """
        Calculate from detailed output report the f1, precision and recall scores per attribute.
        @param var_name: variable name to calculate the scores for.
        @param df: detailed results dataframe with per attribute confusion matrix as outputted by tester.
        @return: f1, precision and recall scores.
        """
        tp = df[var_name + '_tp']
        fp = df[var_name + '_fp']
        tn = df[var_name + '_tn']
        fn = df[var_name + '_fn']
        recall = 0
        precision = 0
        f1 = 0
        if len(np.where(tp + fp + tn + fn != 1)[0]) > 0:
            print(f'{len(np.where(tp + fp + tn + fn != 1)[0])} examples have errors')
        if (tp.sum() + fn.sum()) != 0:
            recall = tp.sum() / (tp.sum() + fn.sum())
        if (tp.sum() + fp.sum()) != 0:
            precision = tp.sum() / (tp.sum() + fp.sum())
        if (precision + recall) != 0:
            f1 = 2 * ((precision * recall) / (precision + recall))
        return f1, precision, recall

    def get_model_performance(self, df):
        """
        Generate aggregated performance report per model. The aggregated report provides per attribute f1, precision and
        recall scores.
        @param df: detailed output report as provided by tester test output.
        @return: aggregated results report
        """
        var_names = ['_'.join(var.split('_')[:-1]) for var in df.columns if var.endswith('_tp')]
        d_l = []
        for var in var_names:
            d = {}
            d['attribute'] = var
            f1, precision, recall = self.get_f1_precision_recall(var, df)
            d['f1'] = f1
            d['precision'] = precision
            d['recall'] = recall
            d_l.append(d)
        res_df = pd.DataFrame(d_l)
        res_df = res_df.set_index('attribute')
        return res_df

    def test_sigparser(self, ground_truth_path):
        """
        Run current local pipeline version test. Each example from the ground truth dataset will be predicted via the
        pipeline and evaluated for accuracy. The output is saved locally with sigparser prefix and stored as a member of
        the tester instance. The outputs are both detailed and aggregated results dataframes.
        @param ground_truth_path: ground truth labeled file path in segment format or expanded format.
        """
        self.build(ground_truth_path)
        self.test()
        report_df = self.report_df
        report_df = self.clear_df_before_saving(report_df)
        report_df.to_csv('local-pipeline_results22_2_21.csv')
        res_df = self.get_model_performance(report_df)
        res_df.to_csv('local-pipeline_aggregated_results_22_2_21.csv')
        self.add_results_to_report(report_df, res_df, 'local-pipeline')

    def add_results_to_report(self, report_df, res_df, algorithm_name):
        self.reports[algorithm_name] = {'detailed': report_df, 'aggregated': res_df}

    def test_from_file(self, ground_truth_path, prediction_path, algorithm_name=None):
        """
        Run a performance test from file output. The file output is compared with the ground truth to produce detailed
        and aggregated result datasets. Both are saved locally with from file prefix and saved as members of the tester
        object reports attribute.
        @param ground_truth_path: ground truth labeled file path in segment of expanded format.
        @param prediction_path: prediction file path in segment of expanded format.
        """
        if self.dataset is None:
            self.build(ground_truth_path)
        if algorithm_name is None:
            algorithm_name = os.path.basename(prediction_path)
        self.test(prediction_path=prediction_path)
        report_df = self.report_df
        report_df = self.clear_df_before_saving(report_df)
        report_df.to_csv(algorithm_name + '_results.csv')
        res_df = self.get_model_performance(report_df)
        res_df.to_csv(algorithm_name + '_aggregated_results.csv')

        self.add_results_to_report(report_df, res_df, algorithm_name)

    def generate_comparative_results(self):
        """
        Use the tester object reports to generate a comparative report between 2 versions. The output are detailed
        results of both versions as in:
        https://docs.google.com/spreadsheets/d/1rMJHHxAU5E0MVkuk5N2JOIh8qi0LeDr4LMiNmqYu8WE/edit#gid=1045120250
        The results provides a comparative aggregated and detailed reoports for both versions together.
        @return: merged datasets for detailed and aggregated results
        """
        def entry_to_dict(entry, prefix):
            d = {}
            for key, val in entry.items():
                d[prefix + key] = val
            return d
        compared_versions = list(self.reports.keys())
        assert len(compared_versions) == 2, 'comparing more than 2 versions is currently not supported'
        # reference version
        v1 = self.reports[compared_versions[0]]['detailed']
        v1_agg = self.reports[compared_versions[0]]['aggregated']
        # compared version
        v2 = self.reports[compared_versions[1]]['detailed']
        v2_agg = self.reports[compared_versions[1]]['aggregated']
        missing_entries = 0  # count missing entries in compared version
        d_l = []
        for i in range(len(v1)):
            v1_entry = v1.iloc[i]
            sig = v1_entry['text']
            match_i = self.dataset.find_signature_in_df(sig, v2, sig_colname='text')
            d = entry_to_dict(v1_entry, compared_versions[0] + '_')
            if match_i:
                v2_entry = v2.iloc[match_i]
                d.update(entry_to_dict(v2_entry, compared_versions[1] + '_'))
            else:
                missing_entries += 1
            d_l.append(d)
        print(f'total missing entries in compared version file: {missing_entries}')
        merged_df = pd.DataFrame(d_l)
        merged_df.to_csv('merged.csv')
        merged_agg = v1_agg.join(v2_agg, on='attribute', lsuffix='_'+compared_versions[0],
                                 rsuffix='_'+compared_versions[1])
        merged_agg.to_csv('merged_agg.csv')
        return merged_df, merged_agg


# usage example for comparing pipeline and file outputs. Note each step can be run independently- so if you wish only to
# test you own local version- just run: tester.test_sigparsr
if __name__ == '__main__':
    ground_truth_path = '/Users/doron.ariav/Desktop/sigparser-dataset/Clean_test_set_10K - Clean_test_set.csv'
    sigparser_file = '/Users/doron.ariav/Desktop/sigparser-dataset/Clean_test_set_10K - 27.12.2020-sigparser.csv'
    cogilex_file = '/Users/doron.ariav/Desktop/sigparser-dataset/Clean_test_set_10K - Parsed with cogilex.csv'
    tester = generatedClass()
    # tester.test_sigparser(ground_truth_path)  # the results are now in tester.reports['sigparser']
    tester.test_from_file(ground_truth_path, sigparser_file)
    tester.test_from_file(ground_truth_path, cogilex_file, algorithm_name='cogilex')  # the results are now in tester.reports['cogilex']
    merged_df, merged_agg_df = tester.generate_comparative_results()  # results are also saved locally (for all steps)

    tester.reports.keys()



from generatedDirectory.pipeline import generatedProjectNamePipeline
from gaiaframework.testable.dataset_base import DatasetBase
import pathlib

class generatedClass(DatasetBase):
    
    @staticmethod
    def get_defaults():
        dir_path = pathlib.Path(__file__).parent.absolute()

        cfg = {}
        cfg['use_local_storage']   = True
        cfg['local_storage_path']  = dir_path/'tests'
        cfg['remote_storage_path'] = ''
        cfg['dataset_name']        = 'annotated_set_round_a_sample.csv'
        return cfg

    def __init__(self, name:str="generatedClassName", path='', load_samples=False,  **kwargs):
        self.pipeline = generatedProjectNamePipeline.getInstance()
        DatasetBase.__init__(self, name, path, False, load_samples, generatedClass.get_defaults(), **kwargs)

    def init_path(self, path=''):
        if not path:
            path = self.local_storage_path/self.dataset_name if self.use_local_storage else self.remote_storage_path/self.dataset_name
        self.path = path

    def set_model_prediction(self):
        # TODO: figure out how to neutralize pre and post processing
        self.pipeline.__dir__()

        raise NotImplementedError


    def set_pipeline_prediction(self):

        raise NotImplementedError

    def set_df_prediction(self, pred_df):

        raise NotImplementedError



    # def get_data_sample_from_row(self, row, i=None):
    #     text               = row.iloc[0]
    #     gt_segments        = row.iloc[1]
    #     # print(text)
    #     gt_segments_dict   = literal_eval(gt_segments) if type(gt_segments) == str else gt_segments
    #     return DatasetSample(x=text, y=gt_segments_dict)



from gaiaframework.testable.reporter_base import ReporterBase
import pathlib
import os


class generatedClass(ReporterBase):
    @staticmethod
    def get_defaults():
        dir_path = pathlib.Path(__file__).parent.absolute()
        reports_path = dir_path / '../reports'
        cfg = {}
        cfg['reports_path'] = reports_path
        cfg['report_header'] = 'Analysis'
        return cfg

    def __init__(self, name:str="generatedClassName", **kwargs):
        ReporterBase.__init__(self, name, generatedClass.get_defaults(),  **kwargs)

    def init_analysis_dict(self):
        d = {
            'confusion_matrix': [
                'name_prefix',
                'name_first',
                'name_middle',
                'name_last',
                'name_suffix',
                'name_credentials',
                'tc_title',
                'tc_company',
                'connections_phones',
                'connections_mobiles',
                'connections_faxes',
                'connections_urls',
                'connections_emails',
                'location_address',
                'location_city',
                'location_state',
                'location_zip',
                'location_country'
            ],
            'f1_precision_recall': [
                'name_prefix',
                'name_first',
                'name_middle',
                'name_last',
                'name_suffix',
                'name_credentials',
                'tc_title',
                'tc_company',
                'connections_phones',
                'connections_mobiles',
                'connections_faxes',
                'connections_urls',
                'connections_emails',
                'location_address',
                'location_city',
                'location_state',
                'location_zip',
                'location_country'
            ],
            # 'intersection_over_union': [
            #     'name_prefix',
            #     'name_first',
            #     'name_middle',
            #     'name_last',
            #     'name_suffix',
            #     'name_credentials',
            #     'tc_title',
            #     'tc_company',
            #     'connections_phones',
            #     'connections_mobiles',
            #     'connections_faxes',
            #     'connections_urls',
            #     'connections_emails',
            #     'location_address',
            #     'location_city',
            #     'location_state',
            #     'location_zip',
            #     'location_country'
            #     ]
        }
        self.analysis_dict = d
        return NotImplementedError

    def get_report_path(self, path = ''):
        reports_path = ''
        if path:
            reports_path = pathlib.Path(path)
        else:
            reports_path = ReporterBase.get_report_path(self)
        report_name  = 'test_' + self.F.get_date_time(short=True)
        report_path  = reports_path + '/' + report_name
        return report_path

    def get_report_header(self, header = ''):
        return header if header else self.report_header

    def init_report_dirs(self, path=''):
        report_path = self.get_report_path(path)
        image_path = report_path + '/' + 'images'
        self.image_dir = image_path
        if not os.path.exists(report_path):
            os.mkdir(report_path)
        if not os.path.exists(image_path):
            os.mkdir(image_path)



import os
import openai
import requests
import json
import tiktoken

class ChatGptService:
    name = 'ChatGpt'
    api_key = None
    def __init__(self, api_key=None):
        self.api_key = api_key

    def run_chat(self, prompt, history=[], stream=False, temperature=0, role='user', model='gpt-3.5-turbo'):
        message = [
            {
                'role': role,
                'content': prompt
            }
        ]
        messages = history + message
        response = openai.ChatCompletion.create(
            api_key=self.api_key,
            model=model,
            messages=messages,
            temperature=temperature,
            stream=stream
        )
        return response

    def run_completion(self, prompt, max_tokens, stream=False, temperature=0, engine="text-davinci-003"):
        completion = openai.Completion.create(
            api_key=self.api_key,
            engine=engine,
            prompt=prompt,
            max_tokens=max_tokens,
            temperature=temperature,
            n=1,
            stream=stream
        )
        return completion

    def get_tokens(self, prompt):
        enc = tiktoken.encoding_for_model("gpt-3.5-turbo")
        tokens = enc.encode(prompt)
        num_tokens = len(tokens)
        return num_tokens

    def get_price(self, tokens):
        price = tokens * (0.002 / 1000)
        return price
    def calculate_price(self, prompt):
        num_tokens = self.get_tokens(prompt)
        return self.get_price(num_tokens)

    def get_price2(self, prompt, model="gpt-3.5-turbo"):
        """Returns the number of tokens used by a list of messages."""
        messages = [prompt]
        try:
            encoding = tiktoken.encoding_for_model(model)
        except KeyError:
            encoding = tiktoken.get_encoding("cl100k_base")
        if model == "gpt-3.5-turbo":  # note: future models may deviate from this
            num_tokens = 0
            for message in messages:
                num_tokens += 4  # every message follows <im_start>{role/name}\n{content}<im_end>\n
                for key, value in message.items():
                    num_tokens += len(encoding.encode(value))
                    if key == "name":  # if there's a name, the role is omitted
                        num_tokens += -1  # role is always required and always 1 token
            num_tokens += 2  # every reply is primed with <im_start>assistant
            print('num_tokens', num_tokens)
            price = num_tokens * (0.002 / 1000)
            return price
        else:
            raise NotImplementedError(f"""num_tokens_from_messages() is not presently implemented for model {model}.
    See https://github.com/openai/openai-python/blob/main/chatml.md for information on how messages are converted to tok""")


import json
from typing import List, Optional
from pydantic import validator
from pydantic.main import BaseModel


class TestInputRow(BaseModel):
    data: str
    id: int
    raw_id: Optional[int] = -1

    @validator('data')
    def data_must_be_valid_json(cls, v):
        try:
            json.loads(v)
        except:
            raise ValueError('data must be a valid json string')
        return v


class TestInputRequest(BaseModel):
    truth_dataset_id: int = -1
    model_type_id: int = -1
    rows: List[TestInputRow]



from typing import Union, List

import logging
import sys
from pythonjsonlogger import jsonlogger

root = logging.getLogger()
root.setLevel(logging.INFO)

handler = logging.StreamHandler(sys.stdout)
handler.setLevel(logging.INFO)
formatter = jsonlogger.JsonFormatter()
handler.setFormatter(formatter)
root.addHandler(handler)

logger = logging.getLogger(__name__)

import os
import json
import uvicorn
from fastapi import FastAPI, Depends, Header, HTTPException
from fastapi.responses import StreamingResponse
from fastapi.middleware.cors import CORSMiddleware
from starlette.requests import Request

from pipeline.pipeline import generatedProjectNamePipeline
from pipeline.schema.inputs import generatedProjectNameInputs
from pipeline.schema.outputs import generatedProjectNameOutputs
from tester.test_schema.test_input import TestInputRequest
from tester.test_schema.test_output import TestOutputResponse
from tester.general_tester import generatedProjectNameGeneralTester


async def get_token_header(x_token: str = Header(...)):
    if x_token != "fake-super-secret-token": # todo: please replace with realistic token
        raise HTTPException(status_code=400, detail="X-Token header invalid")


async def verify_env():
    # Don't allow usage of an endpoint on production environment
    if os.environ.get('SPRING_PROFILES_ACTIVE') == 'production':
        raise HTTPException(status_code=404, detail="Endpoint not available")


def load_allowed_cors_origins():
    allowed_origins = []
    cors_origin_list_path = 'server/cors_allowed_origins.json'
    if os.path.exists(cors_origin_list_path):
        with open(cors_origin_list_path) as f:
            allowed_origins = json.load(f)

    return allowed_origins


app = FastAPI()

origins = load_allowed_cors_origins()
methods = ["*"]
headers = ["*"]
credentials = True

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=credentials,
    allow_methods=methods,
    allow_headers=headers,
)

pipeline = generatedProjectNamePipeline()
# from server.pool import InstancePool
# pipelines = [generatedProjectNamePipeline() for i in range(3)]
# app.pipelines = InstancePool(pipelines)


@app.post('/predict', dependencies=[Depends(get_token_header)])
def predict(body: generatedProjectNameInputs, request: Request) -> List[generatedProjectNameOutputs]:
    '''
    predict api
    '''
    data = body.dict()
    # logger.info({"message": "predict invoked", "data": json.dumps(data)})
    # print('data', data)

    # call model here
    try:
        output: generatedProjectNameOutputs = pipeline.execute(**data)
        logger.info("INFO predict invoked", extra={"input": data, "output": output.dict()})
    except Exception as ex:
        logger.exception(msg=data, exc_info=True)
        output = {'error': {'request': ex}}
        logger.error("ERROR predict invoked", extra={"input": data, "output": output})
    # call model here
    return output


@app.post('/parse', dependencies=[Depends(get_token_header)])
def parse(body: generatedProjectNameInputs, request: Request) -> List[generatedProjectNameOutputs]:
    '''
    parse api
    '''
    data = body.dict()
    # logger.info({"message": "parse invoked", "data": json.dumps(data)})
    # print('data', data)

    # call model here
    try:
        output: generatedProjectNameOutputs = pipeline.execute(**data)
        logger.info("INFO parse invoked", extra={"input": data, "output": output.dict()})
    except Exception as ex:
        logger.exception(msg=data, exc_info=True)
        output = {'error': {'request': ex}}
        logger.error("ERROR parse invoked", extra={"input": data, "output": output})
    # call model here
    return output


# Fetch the proper tester for the project
model_tester = generatedProjectNameGeneralTester()

@app.post('/chat', dependencies=[Depends(get_token_header)])
async def chat(body: generatedProjectNameInputs, request: Request) -> StreamingResponse:
    '''
    parse api
    '''
    data = body.dict()
    # logger.info({"message": "parse invoked", "data": json.dumps(data)})
    # print('data', data)

    # call model here
    output = pipeline.execute_stream(**data)
    async def stream_response():
        count = 0
        final_text = ''
        async for chunk in output:
            count += 1
            final_text = chunk['text']
            yield final_text

    return StreamingResponse(stream_response())

@app.post('/test', dependencies=[Depends(get_token_header), Depends(verify_env)], include_in_schema=False)
def test(input_request: TestInputRequest) -> List[TestOutputResponse]:
    """
    This is the main endpoint for testing. Every project starts with a mock response service, in order to enable
    DSP integration, but as soon as a working testing functions are ready, please replace the mock service with the
    real one.

    Parameters:
        input_request: Request as received from the DSP
    Returns:
        A list of responses, to be sent back to the DSP
    """
    response = model_tester.create_mock_response(input_request)

    """    
    model_tester.save_meta_data(input_request)
    response = []
    # Until this model will support batches, work sequentially. Once it does, pass the full list of rows.
    for request_row in input_request.rows:
        input_list = [request_row]
        tester_output = model_tester.test_batch(input_list)
        response.append(tester_output[0])
    """
    return response


@app.get("/livenessprobe")
def liveness_probe():
    return {"alive": True}


@app.get("/readinessprobe")
def readiness_probe():
    return {"ready": True}


if __name__ == '__main__':
    uvicorn.run(app, host='0.0.0.0', port=8080)




import os
import cohere

class CoHereService:
    name = 'CoHere'
    co = None
    def __init__(self, api_key=None):
        key = api_key
        self.co = cohere.Client(key)
    # @staticmethod
    # def run_chat(prompt, stream=False, temperature=0, role='user', model='gpt-3.5-turbo'):
    #     response = openai.ChatCompletion.create(
    #         model=model,
    #         messages=[
    #             {
    #                 'role': role,
    #                 'content': prompt
    #             }
    #         ],
    #         temperature=temperature,
    #         stream=stream
    #     )
    #     return response

    def run_completion(self, prompt, max_tokens, stream=False, temperature=0.9, model='command-xlarge-nightly'):
        prediction = self.co.generate(
            model=model,
            prompt=prompt,
            max_tokens=max_tokens,
            stream=stream,
            temperature=temperature,
            k=0,
            p=0.75,
            return_likelihoods="NONE",
            stop_sequences=[]
            # frequency_penalty=0.1,
            # presence_penalty=0,
            # stop_sequences=["--"]
        )
        return prediction

    def get_tokens(self, prompt):
        response = self.co.tokenize(
            text=prompt,
        )
        tokens = response.tokens
        token_strings = response.token_strings
        num_tokens = len(tokens)
        return num_tokens

    def get_price(self, tokens):
        price = tokens * (2.5/1000)
        return price
    def calculate_price(self, prompt):
        num_tokens = self.get_tokens(prompt)
        return self.get_price(num_tokens)
    def calc_tokens(self):
        response = self.co.tokenize(
            text='tokenize me! :D',
        )



from locust import HttpUser, task
import pandas as pd
import numpy as np

from server.token_generator import tokenGenerator

t = tokenGenerator()
jwtToken = t.generateToken('STG_Conf')
global_headers = {
    'x-token': 'fake-super-secret-token',
    'Authorization': 'Bearer ' + jwtToken
}

class payloadUser(HttpUser):

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.df = pd.read_csv("sample.csv", sep='\t')
		self.df.replace(np.nan, '', regex=True)

	def get_random_row(self):
		s = self.df.sample().to_dict('list')
		d = dict((k.lower(), v[0]) for k, v in s.items())
		return d

	@task
	def get_diagnosis_with_valid_payload(self):
		headers = {"Content-Type": "application/json; charset=UTF-8", **global_headers}
		payload = self.get_random_row()
		self.client.post("/predict", json=payload, headers=headers)


# how to run
# --host - Host to load test in the following format: http://example.com
# --users - Number of concurrent Locust users
# --spawn-rate - The rate per second in which users are spawned
# locust -f locustfile.py --host={your-host} --users={number-of-users} --spawn-rate={your-spawn-rate}



from gaiaframework.testable.trainer_base import TrainerBase

class generatedClass(TrainerBase):
    @staticmethod
    def get_defaults():
        cfg = {}
        return cfg

    def __init__(self, name:str="generatedClassName",  **kwargs):
        TrainerBase.__init__(self, name, generatedClass.get_defaults(),  **kwargs)
        self.ready = False



# if __name__ == '__main__':



import yaml
from typing import Dict
import os

__location__ = os.path.realpath(
    os.path.join(os.getcwd(), os.path.dirname(__file__)))

from gaiaframework.pipeline import PipelineBase
from gaiaframework.config import ZIDSConfig
from gaiaframework.shared import SharedArtifacts
from gaiaframework.utils.registry import Registry
from generatedDirectory.preprocessor import generatedProjectNamePreprocessor
from generatedDirectory.predictors import generatedProjectNamePredictors
from generatedDirectory.forcers import generatedProjectNameForcers
from generatedDirectory.postprocessor import generatedProjectNamePostprocessor

class generatedClass(PipelineBase):
    @staticmethod
    def getInstance(name = 'generatedDirectory', **kwargs):
        # generatedDirectory =  Registry.initialize(name, None, generatedClass)
        # generatedDirectory.cfg.update(generatedDirectory.name, kwargs)
        return generatedClass

    def __init__(self, config_path: str = 'config.yaml', **kwargs):
        super(generatedClass, self).__init__(name="generatedClassName", config_path=config_path,**kwargs)
    def __call__(self, url: str = '', html: str = '', link_text: str = '', **kwargs):
        return PipelineBase.__call__(self, url=url, html=html, link_text=link_text, **kwargs)

    def init_preprocessor(self, config: ZIDSConfig, shared_object: SharedArtifacts):
        self.preprocess = generatedProjectNamePreprocessor(config=config, shared_artifacts=shared_object)
        return self.preprocess

    def init_predictors(self, config: ZIDSConfig, shared_object: SharedArtifacts) -> None:
        self.predictors = generatedProjectNamePredictors(config=config, shared_artifacts=shared_object)
        raise NotImplementedError

    def init_forcers(self, config: ZIDSConfig, shared_object: SharedArtifacts) -> None:
        self.forcers = generatedProjectNameForcers(config=config, shared_artifacts=shared_object)
        raise NotImplementedError

    def init_postprocessor(self, config: ZIDSConfig, shared_object: SharedArtifacts = None):
        self.postprocess = generatedProjectNamePostprocessor(config=config, shared_artifacts=shared_object)
        return self.postprocess

    @staticmethod
    def read_config(config_path: str = '') -> Dict[str, ZIDSConfig]:
        """
        read configuration from yaml file
        :param config_path: configuration path
        :return:
        """
        with open(os.path.join(__location__, config_path), 'r') as yaml_file:
            cfg = yaml.load(yaml_file, Loader=yaml.FullLoader)
        return {component: ZIDSConfig(cfg[component])
                for component in ['pipeline', 'preprocessor', 'labelizer', 'postprocessor']}



import multiprocessing
bind = "0.0.0.0:8080"
workers = multiprocessing.cpu_count() * 2 + 1
timeout = 60 # Workers silent for more than this many seconds are killed and restarted
graceful_timeout = 60 # After receiving a restart signal, workers have this much time to finish serving requests
worker_class = 'uvicorn.workers.UvicornWorker'

# Gunicorn heartbeat system defaults to /tmp, but Gunicorn FAQ recommends to use a folder that is stored in memory
worker_tmp_dir = "/dev/shm"



import os
import logging as log
from shutil import rmtree
from setuptools import find_packages, setup
from distutils.command.clean import clean
from distutils.cmd import Command

package_name = 'gaia-lib'
version = '0.0.1'
with open('requirements.txt') as f:
    required = f.read().splitlines()


class CleanCommand(clean):
    """
    Custom implementation of ``clean`` setuptools command."""

    def run(self):
        """After calling the super class implementation, this function removes
        the dist directory if it exists."""
        super().run()
        if os.path.exists(".eggs"):
            rmtree(".eggs")
        if os.path.exists(f"{package_name.replace('-', '_')}.egg-info"):
            rmtree(f"{package_name.replace('-', '_')}.egg-info")
        if os.path.exists("dist"):
            rmtree("dist")

def package_files(directory):
    paths = []
    for (path, directories, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join('..', path, filename))
    return paths


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()


dsp_files = package_files('gaia/cli/tester/dsp')
extra_files = package_files('gaia/cli/tester/deploy_files')
cloud_eval_extra_files = package_files('gaia/cli/tester/cloud_eval')

setup(
    name=package_name,
    py_modules=['gaia_cli'],

    packages=find_packages(include=['gaia*']),
    package_data={'': ['config.json', 'cors_allowed_origins.json', '.gitignore'] + extra_files + cloud_eval_extra_files + dsp_files},
    entry_points='''
        [console_scripts]
        gaia-cli=gaia_cli:cli
    ''',
    version=version,
    description='Gaia SDK library',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    # url='http://pypi.python.org/pypi/PackageName/',
    author='oribrau@gmail.com',
    license='MIT',
    install_requires=required,
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    test_suite='tests',
    cmdclass={'clean': CleanCommand}
)

# commands
# for test -  python setup.py pytest
# for build wheel -  python setup.py bdist_wheel
# for source dist -  python setup.py sdist
# for build -  python setup.py build
# for install -  python setup.py install
# for uninstall - python -m pip uninstall gaia-sdk
# for install - python -m pip install dist/gaia-sdk-0.1.0-py3-none-any.whl

# deploy to PyPI
# delete dist and build folders
# python setup.py bdist_wheel
# python setup.py sdist
# python setup.py build
# twine upload dist/*
'''
    use
    1. python setup.py install
    2. gaia-cli g model new_model_name
    3. twine check dist/*
    4. twine upload --repository-url https://pypi.org/legacy/ dist/*
    4. twine upload dist/*
    
    pip install gaia-sdk --index-url https://pypi.org/simple
    
    how to use
    
    pip install gaia-sdk
    
    gaia-cli generate project my-new-model
'''



import asyncio
import csv
import json
from gaia.base.vendors.vendors_manager import VendorsManager
from gaia.base.vendors.chat_gpt import ChatGptService

class Gaia:
    debug=False
    api_token=None
    api_keys: dict = {}
    def __init__(self, api_token, api_keys={}, debug=False):
        self.api_token = api_token
        self.api_keys = api_keys
        self.debug = debug

    def optimize_prompt(self, prompt):
        # TODO run optimize prompt with chat gpt
        c = VendorsManager(debug=self.debug)
        start = '<'
        end = '>'
        param = prompt[prompt.find(start) + len(start):prompt.rfind(end)]
        p = f"please optimize this prompt to get better results for any {param}: {prompt}" \
            f", provide only the final suggested prompt"
        prediction = c.chatGptService.run_chat(p, [], False)
        text = prediction.choices[0].message.content
        return text

    def load_data(self, csv_path):
        csvfile = open(csv_path, 'r', encoding='utf-8')
        # print('csvfile', csvfile)
        csvReader = csv.DictReader(csvfile)
        # print('csvReader', csvReader)
        data = []
        for row in csvReader:
            clean_row = {}
            for k, v in row.items():
                if k:
                    clean_row[k] = ("" if v == None else v)
            # clean_row = {k: ("" if v == None else v) for k, v in row.items()}
            data.append(clean_row)
        return data

    def run_vendors(self, vendors=None, data=[]):
        c = VendorsManager(vendors=vendors, api_keys=self.api_keys, debug=self.debug)
        asyncio.run(c.run_all_parallel(data))
        data = {'vendors_results': c.vendors_results, 'total_took': c.total_took, 'compared_vendors': c.compared_vendors}
        return data

    def save_csv(self, csv_path=None, data=[]):
        csvfile = open(csv_path, 'w', encoding='utf-8', newline="")
        first_row = self.flatten_dict(data[0])
        headers = first_row.keys()
        writer = csv.DictWriter(csvfile,fieldnames=headers)
        writer.writeheader()
        for row in data:
            writer.writerow(self.flatten_dict(row))
        return csv_path

    def flatten_dict(self, dd, separator='_', prefix=''):
        return {prefix + separator + k if prefix else k: v
            for kk, vv in dd.items()
            for k, v in self.flatten_dict(vv, separator, kk).items()
            } if isinstance(dd, dict) else {prefix: dd}

    def get_available_vendors(self):
        return VendorsManager.available_vendors


import multiprocessing

bind = "0.0.0.0:8080"
workers = multiprocessing.cpu_count() + 1  # not much IO in model prediction processing, so maintaining 1:1 process:cpu ratio, plus 1 for standby
timeout = 120  # Workers silent for more than this many seconds are killed and restarted
graceful_timeout = 120  # After receiving a restart signal, workers have this much time to finish serving requests
worker_class = 'uvicorn.workers.UvicornWorker'
worker_tmp_dir = "/mnt/mem"
daemon = True


from gaiaframework.base.common.component import ZIDS_Component

##
# @file
# @brief Forcer class, implements ZIDS_Component base class.
class generatedClass(ZIDS_Component):
    """! Forcer class, implements ZIDS_Component base class.

        This class was added with two main goals, to force results when:

        1. If you know certain data doesn't need to go through the model and would like to force results.
        2. Not satisfied by some results from the model and would like to force another result.

        No examples yet.
    """
    def __init__(self, artifacts=None) -> None:
        """! generatedClass (Forcer) class initializer."""

        ##
        # @hidecallgraph @hidecallergraph
        super().__init__(artifacts)



from cli.allow_commands import AllowCommands
from cli.preprocessor_template import generatedClass
from cli.labelizer_template import generatedClass
from cli.postprocessor_template import generatedClass
from cli.pipeline_template import generatedClass




from typing import List, Any
from gaiaframework.base.common.component import ZIDS_Component

##
# @file
# @brief Predictor class, this is where the model goes.

class generatedClass(ZIDS_Component):
    """! generatedClass class (Predictor) implements ZIDS_Component base class.

    This is the second stage of the pipeline and its main goal is to pass our dataset to the model.
    Basically our model goes in this class.
    """

    def __init__(self, artifacts=None) -> None:
        """generatedClass class (Postprocessor) initializer

        Args:
            artifacts: Shared artifacts instance.
        """

        ##
        # @hidecallgraph @hidecallergraph
        super().__init__(artifacts)

    def execute(self, predictables: List[Any], **kwargs) -> List[Any]:
        """! generatedClass.execute (Predictor) - Run the data against the model in this method.

        Args:
            predictables:List[Any]: List of predictable objects received from Preprocess.
            **kwargs: Dictionary with additional variables, if necessary.
        Returns:
            List[ZIDS_Predictable]: List of predictable objects.
        """

        ##
        # For example:
        # @code{.py}
        #     def execute(self, predictables: List[Any], **kwargs) -> List[Any]:
        #         features = [list(vars(p.features).values()) for p in predictables]
        #
        #         results = self.artifacts.rf_model.predict_proba(features)
        #
        #         idx = 0
        #         for p in predictables:
        #             p.prob = results[idx][-1]
        #             idx += 1
        #         return predictables
        # @endcode

        ##
        # another example:
        # @code
        #     def execute(self, predictables: List[Any], **kwargs) -> List[Any]:
        #         df = pd.DataFrame([predictables.dict()])
        #         df = self.text_data_preparation(df)
        #         data = self.encode_batch(df['text'].tolist())
        #
        #         bert_output = self.artifacts.model(data)
        #
        #         probas = list(tf.nn.softmax(bert_output.logits).numpy())
        #         return probas
        # @endcode

        return predictables



#!/usr/bin/env python
# coding: utf-8

import json
from gaiaframework.base.pipeline.artifacts.shared_artifacts import ZIDS_SharedArtifacts

class generatedClass(ZIDS_SharedArtifacts):

    def __init__(self) -> None:
        super().__init__()

    def extend_load_file_type(self, file_type, path, absolute_path, name):
        if absolute_path:
            if file_type == 'your-file-type':
                with open(absolute_path) as json_file:
                    setattr(self, name, json.load(json_file))



"""! @brief Pipeline entry point."""
import asyncio
from pipeline.pipeline import generatedProjectNamePipeline


##
# @file
#
# @brief Pipeline entry point.
if __name__ == '__main__':
    ## Creates an instance of the pipeline
    p = generatedProjectNamePipeline()

    ## This method executes the pipeline, dataset and required params can be loaded here.\n
    # Examples:
    # @code
    # data = {}
    # for file in data_files:
    #     with open(file) as f:
    #         data = load_json(f)
    #
    # output = p.execute(**data)
    # @endcode
    # or
    # @code
    # p.execute(data=mydata)
    # @endcode
    # or
    # @code
    # sig = Signature(uid=1, text=signature, hints=None)
    # output = p.execute(signatures=[sig])
    # @endcode
    # Define data passed through the execute method in the schema/<my-new-project>Inputs.py class, for example:
    # @code
    # class <my-new-project>Inputs(BaseModel):
    #     signatures: List[Signature]
    # @endcode
    # or
    # @code
    # class <my-new-project>Inputs(BaseModel):
    #     data = {}
    # @endcode

    output = p.execute()
    print(output)


    # for streaming
    # output = p.execute_stream()
    # async def stream_response():
    #     count = 0
    #     final_text = ''
    #     async for chunk in output:
    #         count += 1
    #         final_text = chunk['text']
    #         print('final_text', final_text)
    #
    #
    # loop = asyncio.get_event_loop()
    # loop.run_until_complete(stream_response())



from typing import Any
import requests
from googleapiclient import discovery
from oauth2client.client import GoogleCredentials
import base64
import json
from datetime import datetime, timedelta


class tokenGenerator():
	# account example - original_project_name@gaia-stg-apps.iam.gserviceaccount.com
	# audience example - 1042519062770-rghbs1ied8kno7l00mlnko5r7sgjea2o.apps.googleusercontent.com
	conf = {
		"STG_Conf": {
			"account": '', # need to get from dev ops
			"host": 'gaia-stg-apps',
			"target_audience": '1042519062770-rghbs1ied8kno7l00mlnko5r7sgjea2o.apps.googleusercontent.com', # need to get from dev ops
			"iap": True,
		},
		"PRD_Conf": {
			"account": '',  # need to get from dev ops
			"host": 'gaia-prd-apps',
			"target_audience": '484284403458-lpmc6edp4b6rd2383dgpkiq5a9ngkpbc.apps.googleusercontent.com',  # need to get from dev ops
			"iap": True,
		}
	}
	tokens = {
		"STG_Conf": '',
		"PRD_Conf": '',
	}

	def generateToken(self, type):
		if self.isExpired(type):
			self.resetToken(type)
		if not type or not self.conf[type]:
			return ''
		conf: Any = self.conf[type]
		if not conf['account']:
			print('please specify account')
			return ''
		if not conf['target_audience']:
			print('please specify target audience')
			return ''
		if not self.tokens[type]:
			credentials = self.authorize()
			token = self.getToken(credentials, conf)
			if self.conf[type]['iap']:
				token = self.buildIdToken(credentials, token, conf)
			self.tokens[type] = token

		return self.tokens[type]

	def authorize(self):
		# return google.auth.default(
		# 	scopes=['https://www.googleapis.com/auth/cloud-platform'])
		return GoogleCredentials.get_application_default()

	def getToken(self, credentials, conf):
		client_id = credentials.client_id
		client_secret = credentials.client_secret
		aud = credentials.token_uri
		iat = int(datetime.now().timestamp())
		one_hours_from_now = datetime.now() + timedelta(hours=1)
		exp = int(one_hours_from_now.timestamp())
		service = discovery.build('iam', 'v1', credentials=credentials)

		name = f'projects/{conf["host"]}/serviceAccounts/{conf["account"]}'
		data ={
			'url': f'https://iam.googleapis.com/v1/projects/-/serviceAccounts/{conf["account"]}:signJwt',
			'data': {
				'payload': json.dumps({"aud": aud,"target_audience": conf["target_audience"],"sub": conf["account"],"iss": conf["account"],"iat": iat, "exp": exp})
			}
		}
		request = service.projects().serviceAccounts().signJwt(name=name, body={'payload': data["data"]["payload"]})
		response = request.execute()
		return response['signedJwt']

	def buildIdToken(self, client, jwtToken, conf):
		try:
			data = {
				"method": 'post',
				"url": 'https://www.googleapis.com/oauth2/v4/token',
				"data": {
					'grant_type': "urn:ietf:params:oauth:grant-type:jwt-bearer",
					'assertion': jwtToken
				}
			}
			# print('data', data)
			result = requests.post(data['url'], data['data'])
			if result and result.content:
				parsed = json.loads(result.content)
				print("[refreshTokens] token for host " + conf['host'] + " refreshed successfully")
				return parsed['id_token']
			else:
				print('no data.id_token returned from googleapis:signJwt')
		except Exception as ex:
			print("[refreshTokens] token for host " + conf['host'] + " failed to refresh")
			print(ex)

	def isExpired(self, type):
		isExpired = True
		if self.tokens[type]:
			decoded = self.parse_id_token(self.tokens[type])
			if decoded and int(decoded['exp']) > int(datetime.now().timestamp()):
				isExpired = False
		return isExpired

	def isExpiredToken(self, token):
		isExpired = True
		decoded = self.parse_id_token(token)
		if decoded and int(decoded['exp']) > int(datetime.now().timestamp()):
			isExpired = False
		return isExpired

	def resetToken(self, type):
		self.tokens[type] = ''

	@staticmethod
	def parse_id_token(token: str) -> dict:
		"""
		Parse Google OAuth2.0 id_token payload
		An ID token has the following structure: Base64(JOSE header).Base64(Payload).Base64(Signature)
		"""
		parts = token.split(".")
		if len(parts) != 3:
			raise Exception("Incorrect id token format")
		payload = parts[1]
		padded = payload + "=" * (4 - len(payload) % 4)
		decoded = base64.b64decode(padded)
		return json.loads(decoded)


if __name__ == '__main__':
	t = tokenGenerator()
	t = t.generateToken('STG_Conf')
	print(t)

